/*
 * missile - sim.c
 *
 * Copyright (C) 2015..2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <getopt.h>
#include <pid.h>

#ifndef PACKAGE_VERSION
#define PACKAGE_VERSION "0.0"
#endif

/*
 * missile configuration
 */
static unsigned int num_gyros = 1;
static unsigned int num_small_thrusters;
static unsigned int num_large_thrusters = 1;
static unsigned int armor = 2;
static double target_g = 0.1;
static int verbose;

/* maximal linear and angular velocity, from Maximum Speed Limit */
#define VMAX		1000
#define VAMAX		20

#define SMALL_THRUSTER_FORCE_MAX	12000
#define LARGE_THRUSTER_FORCE_MAX	144000
#define GYRO_MOMENT_MAX			448000

#define SMALL_THRUSTER_MASS		93
#define LARGE_THRUSTER_MASS		721
#define BATTERY_MASS			1040.4
#define GYRO_MASS			613.6
#define WARHEAD_MASS			26.2
#define PROGRAMMABLE_BLOCK_MASS		158.4
#define ANTENNA_MASS			52.2
#define LIGHT_ARMOR_MASS		20

enum {
	Q_X = 0,
	Q_Y,
	Q_Z,
	Q_YAW,
	Q_PITCH,
	Q_ROLL,
	Q_N
};

struct pvector {
	double q[6];
};

#define PVECTOR(X, Y, Z, A, B, C) \
	((struct pvector) { \
		.q = { X, Y, Z, A, B, C } \
	})

#define DEFINE_PVECTOR(name, X, Y, Z, A, B, C) \
	struct pvector name = PVECTOR(X, Y, Z, A, B, C)


struct object {
	const char *id;
	struct pvector im;

	struct pvector x;
	struct pvector v;
};

struct missile {
	struct object object;

	struct pid pitch_pid;
	struct pid yaw_pid;

	struct pvector old_target;

	double energy_consumed;
};

static void pvector_mov(const struct pvector *a, struct pvector *d)
{
	unsigned int i;

	for (i = 0; i < Q_N; i++)
		d->q[i] = a->q[i];
}

static void pvector_add(const struct pvector *a,
		const struct pvector *b, struct pvector *c)
{
	unsigned int i;

	for (i = 0; i < Q_N; i++)
		c->q[i] = a->q[i] + b->q[i];
}

static void pvector_addm(const struct pvector *a,
		const struct pvector *b, double c, struct pvector *d)
{
	unsigned int i;

	for (i = 0; i < Q_N; i++)
		d->q[i] = a->q[i] + c * b->q[i];
}

static void pvector_sub(const struct pvector *a,
		const struct pvector *b, struct pvector *c)
{
	unsigned int i;

	for (i = 0; i < Q_N; i++)
		c->q[i] = a->q[i] - b->q[i];
}

static double pvector_len(const struct pvector *a)
{
	double l = 0;
	unsigned int i;

	for (i = 0; i < Q_YAW; i++)
		l += a->q[i] * a->q[i];
	return sqrt(l);
}

static void pvector_mul_const(const struct pvector *a, double c,
		struct pvector *d)
{
	unsigned int i;

	for (i = 0; i < Q_YAW; i++)
		d->q[i] = c * a->q[i];
}

static void object_sim(struct object *o, const struct pvector *f, double dt)
{
	unsigned int i;
	double v;

	for (i = 0; i < Q_N; i++) {
		o->x.q[i] += o->v.q[i] * dt;
		o->v.q[i] += f->q[i] * dt * o->im.q[i];
	}
	v = pvector_len(&o->v);
	if (fabs(v) > VMAX)
		pvector_mul_const(&o->v, VMAX / fabs(v), &o->v);
	for (i = Q_PITCH; i < Q_N; i++) {
		if (o->v.q[i] > VAMAX)
			o->v.q[i] = VAMAX;
		else if (o->v.q[i] < -VAMAX)
			o->v.q[i] = -VAMAX;
	}
}

static void object_init(struct object *o, const char *id, double mass, double r)
{
	unsigned int i;

	o->id = id;
	for (i = 0; i < Q_N; i++) {
		o->x.q[i] = 0;
		o->v.q[i] = 0;
	}
	for (i = 0; i < Q_YAW; i++)
		o->im.q[i] = 1 / mass;
	for (; i < Q_N; i++)
		o->im.q[i] = 1 / (mass * r * r);
	fprintf(stderr, "%s: im=%f %f %f  %f %f %f\n",
			o->id,
			o->im.q[Q_X], o->im.q[Q_Y], o->im.q[Q_Z],
			o->im.q[Q_YAW], o->im.q[Q_PITCH], o->im.q[Q_ROLL]);
}

static void object_print(struct object *o, FILE *f)
{
	fprintf(f, "%10s: x: %5.3f  y: %5.3f  z: %5.3f   y: %5.1f  p: %5.1f  r: %5.1f\n",
			o->id,
			o->x.q[Q_X], o->x.q[Q_Y], o->x.q[Q_Z],
			180.0 / M_PI * o->x.q[Q_YAW],
			180.0 / M_PI * o->x.q[Q_PITCH],
			180.0 / M_PI * o->x.q[Q_ROLL]);
}

static void rotate_z(const struct pvector *a, double alpha,
		struct pvector *d)
{
	double c_a = cos(alpha);
	double s_a = sin(alpha);
	double x = c_a * a->q[Q_X] - s_a * a->q[Q_Y];
	double y = s_a * a->q[Q_X] + c_a * a->q[Q_Y];
	double z = a->q[Q_Z];

	d->q[Q_X] = x;
	d->q[Q_Y] = y;
	d->q[Q_Z] = z;
}

static void rotate_y(const struct pvector *a, double alpha,
		struct pvector *d)
{
	double c_a = cos(alpha);
	double s_a = sin(alpha);
	double x = c_a * a->q[Q_X] + s_a * a->q[Q_Z];
	double y = a->q[Q_Y];
	double z = -s_a * a->q[Q_X] + c_a * a->q[Q_Z];

	d->q[Q_X] = x;
	d->q[Q_Y] = y;
	d->q[Q_Z] = z;
}

#define MISSILE_FMAX	(num_small_thrusters * SMALL_THRUSTER_FORCE_MAX \
		+ num_large_thrusters * LARGE_THRUSTER_FORCE_MAX)
#define GYRO_MAX	(num_gyros * GYRO_MOMENT_MAX)

static void missile_init(struct missile *m, const char *id, double mass,
		double r)
{
	object_init(&m->object, id, mass, r);

	pid_init(&m->pitch_pid, 100, 0.4, 0.2, 200);
	pid_set_interval(&m->pitch_pid, 0.01);
	pid_init(&m->yaw_pid, 100, 0.4, 0.2, 200);
	pid_set_interval(&m->yaw_pid, 0.01);
	m->old_target = PVECTOR(0, 0, 0, 0, 0, 0);
	m->energy_consumed = 0;

	fprintf(stderr, "missile: %s, weight: %.0f kg, max. load: %.2f g\n",
			id, mass, MISSILE_FMAX / (mass * 9.81));
}

static double norm2(double a, double b)
{
	return sqrt(a * a + b * b);
}

static void missile_sim(struct missile *m, const struct pvector *target_pos,
		double dt)
{
	double dv = MISSILE_FMAX * m->object.im.q[Q_X];
	DEFINE_PVECTOR(f, 0, 0, 0, 0, 0, 0);
	struct pvector error;
	struct pvector v_error;
	struct pvector v;
	struct pvector target;
	struct pvector target_v;
	double distance;
	double v_sp = 100;
	double pitch_e;
	double yaw_e;
	double t;
	double p;
	unsigned int i;

	/* acquire target */
	pvector_sub(target_pos, &m->object.x, &error);
	distance = pvector_len(&error);
	if (distance < 1e-12)
		distance = 1e-12;
	pvector_sub(target_pos, &m->old_target, &target_v);
	pvector_mov(target_pos, &m->old_target);

	v_sp = 1.5 * pvector_len(&target_v) / dt;
	if (v_sp < 100)
		v_sp = 100;

	for (i = 0; i < 10; i++) {
		t = distance / v_sp;
		pvector_addm(target_pos, &target_v, t / dt, &target);

		pvector_sub(&target, &m->object.x, &error);
		distance = pvector_len(&error);
		if (distance < 1e-12)
			distance = 1e-12;
	}

	pvector_mov(&m->object.v, &v);
	pvector_mul_const(&error, v_sp / distance, &v_error);
	pvector_sub(&v_error, &v, &v_error);

	rotate_z(&v_error, -m->object.x.q[Q_YAW], &v_error);
	rotate_y(&v_error, -m->object.x.q[Q_PITCH], &v_error);

	pitch_e = -atan2(v_error.q[Q_Z], norm2(v_error.q[Q_X], v_error.q[Q_Y]));
	yaw_e = -atan2(v_error.q[Q_Y], v_error.q[Q_X]);

	pid_step(&m->pitch_pid, pitch_e / (m->object.im.q[Q_PITCH] * GYRO_MAX));
	f.q[Q_PITCH] = pid_output(&m->pitch_pid);

	pid_step(&m->yaw_pid, yaw_e / (m->object.im.q[Q_YAW] * GYRO_MAX));
	f.q[Q_YAW] = pid_output(&m->yaw_pid);

	f.q[Q_PITCH] *= GYRO_MAX;
	f.q[Q_YAW] *= GYRO_MAX;

	p = 0.1 * pvector_len(&v_error) / (dv * dt);
	p *= v_error.q[Q_X] / pvector_len(&v_error);
	p += 1e-6;
	if (p > 1)
		p = 1;
	f.q[Q_X] = 0;
	if (t < 2) {
		if (v_error.q[Q_X] > 0.0)
			f.q[Q_X] = 1;
	} else {
		if (v_error.q[Q_X] > 0.5 * pvector_len(&v_error))
			f.q[Q_X] = p;
		else if (v_error.q[Q_X] > 0.0)
			f.q[Q_X] = 1e-6;
	}
	m->energy_consumed += f.q[Q_X] * 2.4e6 * dt;

	rotate_y(&f, m->object.x.q[Q_PITCH], &f);
	rotate_z(&f, m->object.x.q[Q_YAW], &f);

	if (verbose >= 2) {
		fprintf(stdout, "EV: %.2f %.2f %.2f  NE: %.2f %.2f %.2f  "
				"TV: %.2f %.2f %.2f  %7.1f p: %.1f\n",
			error.q[Q_X], error.q[Q_Y], error.q[Q_Z],
			v_error.q[Q_X], v_error.q[Q_Y], v_error.q[Q_Z],
			f.q[Q_X], f.q[Q_Y], f.q[Q_Z],
			f.q[Q_PITCH], 180 / M_PI * m->object.x.q[Q_PITCH]);
	}

	pvector_mul_const(&f, MISSILE_FMAX, &f);
	object_sim(&m->object, &f, dt);
}

static void rng_init(void)
{
	int fd;
	long seed = 0;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd == -1) {
		perror("open(/dev/urandom)");
		return;
	}
	if (read(fd, &seed, sizeof(seed)) != sizeof(seed))
		perror("read");
	close(fd);
	srand48(seed);
}

static const char copyright_string[] =
"missile-sim " PACKAGE_VERSION ".\n"
"Copyright (C) 2015..2016  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [--] <id...>\n"
"Options:\n"
"  -g, --gyros=GYROS          set number of gyros\n"
"  -s, --small=SMALL          set number of small thrusters\n"
"  -l, --large=LARGE          set number of large thrusters\n"
"  -h, --help                 show this help\n"
"  -t, --target-g=G-LIMIT     set target g-limit\n"
"  -v, --verbose              be verbose\n"
"  -V, --version              show version\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, name ? "missile-sim" : name);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{"small", 1, NULL, 's'},
		{"gyros", 1, NULL, 'g'},
		{"large", 1, NULL, 'l'},
		{"help", 0, NULL, 'h'},
		{"target-g", 1, NULL, 't'},
		{"verbose", 0, NULL, 'v'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "g:hl:t:s:vV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'g':
			num_gyros = atoi(optarg);
			break;
		case 'h':
			show_usage(argv[0]);
			break;
		case 'l':
			num_large_thrusters = atoi(optarg);
			break;
		case 't':
			target_g = atof(optarg);
			break;
		case 's':
			num_small_thrusters = atoi(optarg);
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			show_version();
			break;
		default:
			fprintf(stderr, "Unknown option: %d.\n", c);
			exit(EXIT_FAILURE);
		}
	}

	while (optind < argc) {
		fprintf(stderr, "warning: unused argument: '%s'.\n",
				argv[optind++]);
	}
}

int main(int argc, char **argv)
{
	struct object target;
	struct missile missile;
	unsigned long i;
	double dt = 0.01;
	double missile_mass = 0;
	double min_dist = 1e10;
	double dist = 1000;
	DEFINE_PVECTOR(f, 0, 0, 0, 0, 0, 0);

	parse_command_line(argc, argv);

	rng_init();
	/*
	BBBBBB
	2222GB
	2222WB
	2222PB
	BBBBBB
	*/

	missile_mass += num_small_thrusters * SMALL_THRUSTER_MASS;
	missile_mass += num_large_thrusters * LARGE_THRUSTER_MASS;
	missile_mass += BATTERY_MASS;
	missile_mass += num_gyros * GYRO_MASS;
	missile_mass += WARHEAD_MASS;
	missile_mass += PROGRAMMABLE_BLOCK_MASS;
	missile_mass += ANTENNA_MASS;

	if (armor == 2)
		missile_mass += (2 * 6 * 5 + 2 * 15) * LIGHT_ARMOR_MASS;
	else if (armor == 1)
		missile_mass += (2 * 3 * 5 + 2 * 9) * LIGHT_ARMOR_MASS;
	object_init(&target, "target", 1, 30);
	missile_init(&missile, "missile", missile_mass, 5);

	target.x.q[Q_X] = 9000;
	target.x.q[Q_Y] = 1000;
	target.x.q[Q_Z] = 1000;

	target.v.q[Q_Z] = 0;
	target.v.q[Q_X] = 0;
	target.v.q[Q_Z] = -80;

	missile.object.x.q[Q_PITCH] = 0;
	missile.object.x.q[Q_YAW] = 0;
	missile.object.v.q[Q_X] = 0;

	for (i = 0; i < 100000; i++) {
		struct pvector d;

#if 1
		if ((i % 500) == 499) {
			double r = drand48();
			if (r < 0.25)
				rotate_y(&target.v, M_PI_2, &f);
			else if (r < 0.5)
				rotate_y(&target.v, -M_PI_2, &f);
			else if (r < 0.75)
				rotate_z(&target.v, M_PI_2, &f);
			else
				rotate_z(&target.v, -M_PI_2, &f);
			pvector_mul_const(&f, target_g * 9.81 / pvector_len(&f), &f);
			pvector_mul_const(&target.v, 50 / pvector_len(&target.v),
					&target.v);
		}
#elif 0
		if ((i % 500) == 499) {
			if (drand48() < 0.5)
				rotate_y(&target.v, M_PI_2, &f);
			else
				rotate_y(&target.v, -M_PI_2, &f);
			pvector_mul_const(&f, target_g * 9.81 / pvector_len(&f), &f);
			pvector_mul_const(&target.v, 50 / pvector_len(&target.v),
					&target.v);
		}
#elif 1
//		if ((i / 200) % 2)
//			rotate_y(&target.v, -M_PI_2, &f);
//		else
			rotate_y(&target.v, M_PI_2, &f);
		pvector_mul_const(&f, 2 * 9.81 / pvector_len(&f), &f);
#endif

//		target.v.q[Q_Z] = 100 * cos(0.04 * 2 * M_PI * i * dt);
//		target.v.q[Q_X] = 100 * sin(0.04 * 2 * M_PI * i * dt);

		if (verbose) {
			fprintf(stdout, "TARGET %ld  %f %f %f  %f %f %f  "
					"%f %f %f  %f %f  %f\n",
					i,
					missile.object.x.q[Q_X],
					missile.object.x.q[Q_Y],
					missile.object.x.q[Q_Z],
					target.x.q[Q_X],
					target.x.q[Q_Y],
					target.x.q[Q_Z],
					dist,
					pvector_len(&missile.object.v),
					180/M_PI * missile.object.x.q[Q_PITCH],
					180/M_PI * missile.pitch_pid.e1,
					180/M_PI * missile.yaw_pid.e1,
					missile.pitch_pid.u
					);
		}

		missile_sim(&missile, &target.x, dt);
		object_sim(&target, &f, dt);

		pvector_sub(&missile.object.x, &target.x, &d);
		dist = pvector_len(&d);
//		if ((dist > 5) && (min_dist < 5) && (dist > min_dist)) {
		if (dist > min_dist + 2) {
			fprintf(stderr, "HIT T=%.1fs, distance %.3f m, velocity: %.3f\n",
					i * dt, min_dist, pvector_len(&missile.object.v));
			fprintf(stderr, "target velocity: %.3f\n",
					pvector_len(&target.v));
			fprintf(stderr, "energy consumed: %.6f MWh\n",
					1e-6/3600 * missile.energy_consumed);
			min_dist = 1000;
			exit(0);
		}
		if (dist < min_dist)
			min_dist = dist;
	}
	return 0;
}
